import types

from ..utils import get_db_manager_from_uri


def _get_config_from_path(filename):
    # See flask.config.Config.from_pyfile()
    d = types.ModuleType("config")
    d.__file__ = filename
    try:
        with open(filename, mode="rb") as config_file:
            exec(compile(config_file.read(), filename, "exec"), d.__dict__)  # noqa: S102
    except OSError as e:
        e.strerror = f"Unable to load configuration file ({e.strerror})"
        raise
    config = {}
    for key in dir(d):
        if key.isupper():
            config[key] = getattr(d, key)
    return config


def get_db_manager_from_config(filename):
    config = _get_config_from_path(filename)
    dburi = config["SQLALCHEMY_DATABASE_URI"]
    return get_db_manager_from_uri(dburi)



from tahrir_api.dbapi import TahrirDatabase

db = TahrirDatabase("backend://badges:badgesareawesome@localhost/badges")

origin = "http://foss.rit.edu/badges"
issuer_name = "FOSS@RIT"
org = "http://foss.rit.edu"
contact = "foss@rit.edu"

issuer_id = db.add_issuer(origin, issuer_name, org, contact)


badge_name = "fossbox"
image = "http://foss.rit.edu/files/fossboxbadge.png"
desc = "Welcome to the FOSSBox. A member is you!"
criteria = "http://foss.rit.edu"

db.add_badge(badge_name, image, desc, criteria, issuer_id)



import pytest

from tahrir_api.dbapi import TahrirDatabase
from tahrir_api.utils import get_db_manager_from_uri


@pytest.fixture
def callback_calls():
    return []


@pytest.fixture
def api(callback_calls, tmp_path):
    def callback(*args, **kwargs):
        callback_calls.append((args, kwargs))

    db_uri = f"sqlite:///{tmp_path.as_posix()}/testdb.db"
    db_api = TahrirDatabase(db_uri, notification_callback=callback)
    db_mgr = get_db_manager_from_uri(db_uri)
    db_mgr.sync()
    return db_api



import datetime

import pytest

from tahrir_api.model import Assertion


now = datetime.datetime.now()
yesterday = now - datetime.timedelta(days=1)
one_week_ago = now - datetime.timedelta(days=7)
one_month_ago = now - datetime.timedelta(weeks=4)


def assert_in(member, container):
    """Just like assertTrue(a in b), but with a nicer default message."""
    if member not in container:
        raise AssertionError(f"{member} not found in {container}")


@pytest.fixture
def test_data(api):
    issuer_id = api.add_issuer("TestOrigin", "TestName", "TestOrg", "TestContact")
    badge_id_1 = api.add_badge(
        "TestBadge1",
        "TestImage",
        "A test badge for doing unit tests",
        "TestCriteria",
        issuer_id,
    )
    badge_id_2 = api.add_badge(
        "TestBadge2",
        "TestImage",
        "A test badge for doing unit tests",
        "TestCriteria",
        issuer_id,
    )
    badge_id_3 = api.add_badge(
        "TestBadge3",
        "TestImage",
        "A test badge for doing unit tests",
        "TestCriteria",
        issuer_id,
    )
    email_1 = "test_1@tester.com"
    api.add_person(email_1)
    email_2 = "test_2@tester.com"
    api.add_person(email_2)
    email_3 = "test_3@tester.com"
    api.add_person(email_3)
    email_4 = "test_4@tester.com"
    api.add_person(email_4)
    return {
        "badge_1": badge_id_1,
        "badge_2": badge_id_2,
        "badge_3": badge_id_3,
        "email_1": email_1,
        "email_2": email_2,
        "email_3": email_3,
        "email_4": email_4,
    }


def test_ranking_simple(api, test_data):
    api.add_assertion(test_data["badge_1"], test_data["email_1"], None)

    api.add_assertion(test_data["badge_1"], test_data["email_4"], None)
    api.add_assertion(test_data["badge_2"], test_data["email_4"], None)
    api.add_assertion(test_data["badge_3"], test_data["email_4"], None)

    person1 = api.get_person("test_1@tester.com")
    person4 = api.get_person("test_4@tester.com")

    assert person1.rank == 2
    assert person4.rank == 1


def test_ranking_tie(api, test_data):
    api.add_assertion(test_data["badge_1"], test_data["email_1"], None)

    api.add_assertion(test_data["badge_1"], test_data["email_2"], None)
    api.add_assertion(test_data["badge_2"], test_data["email_2"], None)

    api.add_assertion(test_data["badge_1"], test_data["email_3"], None)
    api.add_assertion(test_data["badge_2"], test_data["email_3"], None)

    api.add_assertion(test_data["badge_1"], test_data["email_4"], None)
    api.add_assertion(test_data["badge_2"], test_data["email_4"], None)
    api.add_assertion(test_data["badge_3"], test_data["email_4"], None)

    person1 = api.get_person("test_1@tester.com")
    person2 = api.get_person("test_2@tester.com")
    person3 = api.get_person("test_3@tester.com")
    person4 = api.get_person("test_4@tester.com")

    assert person1.rank == 4
    assert person2.rank == 2
    assert person3.rank == 2
    assert person4.rank == 1


def test_ranking_preexisting(api, test_data):
    """Test that rank updating works for pre-existant users"""
    person1 = api.get_person("test_1@tester.com")
    new_assertion1 = Assertion(badge_id=test_data["badge_1"], person_id=person1.id)
    api.session.add(new_assertion1)
    new_assertion2 = Assertion(badge_id=test_data["badge_2"], person_id=person1.id)
    api.session.add(new_assertion2)
    api.session.flush()

    # For persons who existed *before* we added cached ranks, they should
    # have a null-rank.
    assert person1.rank is None

    # But once *anyone* else gets a badge, old ranks should be updated too.
    api.add_assertion(test_data["badge_1"], test_data["email_2"], None)
    assert person1.rank == 1

    person2 = api.get_person("test_2@tester.com")
    assert person2.rank == 2

    # but people with no badges should still be null ranked.
    person3 = api.get_person("test_3@tester.com")
    assert person3.rank is None


def test_ranking_with_time_limits(api, test_data):
    api.add_assertion(test_data["badge_1"], test_data["email_1"], yesterday)

    api.add_assertion(test_data["badge_1"], test_data["email_4"], yesterday)
    api.add_assertion(test_data["badge_2"], test_data["email_4"], one_week_ago)
    api.add_assertion(test_data["badge_3"], test_data["email_4"], one_month_ago)

    person1 = api.get_person("test_1@tester.com")
    person4 = api.get_person("test_4@tester.com")

    epsilon = datetime.timedelta(hours=1)

    results = api._make_leaderboard(yesterday - epsilon, now)
    assert results[person1]["badges"] == 1
    assert results[person4]["badges"] == 1

    results = api._make_leaderboard(one_week_ago - epsilon, now)
    assert results[person1]["badges"] == 1
    assert results[person4]["badges"] == 2

    results = api._make_leaderboard(one_month_ago - epsilon, now)
    assert results[person1]["badges"] == 1
    assert results[person4]["badges"] == 3



"""Create new tables for Quest

Revision ID: e7040e76728
Revises: 508367dcbbb5
Create Date: 2016-09-02 20:51:14.951460

"""

import datetime

import sqlalchemy as sa
from alembic import op


# revision identifiers, used by Alembic.
revision = "e7040e76728"
down_revision = "508367dcbbb5"


def generate_default_id(context):
    return context.current_parameters["name"].lower().replace(" ", "-")


def upgrade():

    op.create_table(
        "team",
        sa.Column("id", sa.Unicode(length=128), nullable=False, default=generate_default_id),
        sa.Column("name", sa.Unicode(length=128), nullable=False),
        sa.Column(
            "created_on",
            sa.DateTime(),
            nullable=False,
            default=datetime.datetime.now,
        ),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("name"),
    )
    op.create_table(
        "series",
        sa.Column("id", sa.Unicode(length=128), nullable=False, default=generate_default_id),
        sa.Column("name", sa.Unicode(length=128), nullable=False),
        sa.Column("description", sa.Unicode(length=128), nullable=False),
        sa.Column(
            "created_on",
            sa.DateTime(),
            nullable=False,
            default=datetime.datetime.now,
        ),
        sa.Column(
            "last_updated",
            sa.DateTime(),
            nullable=False,
            default=datetime.datetime.now,
            onupdate=datetime.datetime.now,
        ),
        sa.Column("tags", sa.Unicode(length=128), nullable=True),
        sa.Column("team_id", sa.Unicode(length=128), nullable=False),
        sa.ForeignKeyConstraint(["team_id"], ["team.id"]),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("name"),
    )
    op.create_table(
        "milestone",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("position", sa.Integer(), nullable=True),
        sa.Column("badge_id", sa.Unicode(length=128), nullable=False),
        sa.Column("series_id", sa.Unicode(length=128), nullable=False),
        sa.ForeignKeyConstraint(["badge_id"], ["badges.id"]),
        sa.ForeignKeyConstraint(["series_id"], ["series.id"]),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("id"),
        sa.UniqueConstraint("position", "badge_id", "series_id"),
    )


def downgrade():

    op.drop_table("milestone")
    op.drop_table("series")
    op.drop_table("team")



import os
import sys

import click

from .utils import get_db_manager_from_config


def usage(argv):
    cmd = os.path.basename(argv[0])
    print(f"usage: {cmd} <config_uri>\n '(example: \"{cmd} development.ini\"'")
    sys.exit(1)


@click.command()
@click.argument("config", type=click.Path(exists=True))
def main(config):
    db_mgr = get_db_manager_from_config(config)
    db_mgr.sync()



"""Add created_by field to invitations.

Revision ID: 24282792d72a
Revises: 5791a2b9fb6a
Create Date: 2013-06-10 15:51:02.288685

"""

import sqlalchemy as sa
from alembic import op


# revision identifiers, used by Alembic.
revision = "24282792d72a"
down_revision = "5791a2b9fb6a"


def upgrade():
    op.add_column(
        "invitations",
        sa.Column("created_by", sa.Unicode(128), sa.ForeignKey("persons.id"), nullable=False),
    )


def downgrade():
    # Downgrade will fail if using SQLite. The transaction will
    # be rolled back as per env.py.
    op.drop_column("invitations", "created_by")



import pytest


@pytest.fixture
def dummy_issuer_id(api):
    return api.add_issuer("TestOrigin", "TestName", "TestOrg", "TestContact")


@pytest.fixture
def dummy_badge_id(api, dummy_issuer_id):
    return api.add_badge(
        "TestBadge",
        "TestImage",
        "A test badge for doing unit tests",
        "TestCriteria",
        dummy_issuer_id,
    )


@pytest.fixture
def dummy_person_id(api):
    return api.add_person("test@tester.com")


def test_add_badges(api, dummy_badge_id):
    assert api.get_badge("testbadge").__str__() == "TestBadge"
    assert api.badge_exists("testbadge") is True


def test_add_team(api):
    api.create_team("TestTeam")

    assert api.team_exists("testteam") is True


def test_add_series(api):
    team_id = api.create_team("TestTeam")

    api.create_series("TestSeries", "A test series", team_id, "test, series")

    assert api.series_exists("testseries") is True


def test_add_milestone(api, dummy_issuer_id):
    team_id = api.create_team("TestTeam")
    series_id = api.create_series("TestSeries", "A test series", team_id, "test, series")

    badge_id_1 = api.add_badge(
        "TestBadge-1",
        "TestImage-2",
        "A test badge for doing 10 unit tests",
        "TestCriteria",
        dummy_issuer_id,
    )

    badge_id_2 = api.add_badge(
        "TestBadge-2",
        "TestImage-2",
        "A test badge for doing 100 unit tests",
        "TestCriteria",
        dummy_issuer_id,
    )

    milestone_id_1 = api.create_milestone(1, badge_id_1, series_id)

    milestone_id_2 = api.create_milestone(2, badge_id_2, series_id)

    assert api.milestone_exists(milestone_id_1) is True
    assert api.milestone_exists(milestone_id_2) is True


def test_add_person(api, dummy_person_id):
    assert api.get_person("test@tester.com").__str__() == "test@tester.com"
    assert api.person_exists("test@tester.com") is True


def test_add_issuer(api, dummy_issuer_id):
    assert api.get_issuer(dummy_issuer_id).__str__() == "TestName"
    assert api.issuer_exists("TestOrigin", "TestName") is True


def test_add_invitation(api, dummy_badge_id, dummy_person_id):
    _id = api.add_invitation(dummy_badge_id, created_by_email="test@tester.com")

    assert api.invitation_exists(_id)
    invitation = api.get_invitation(_id)
    assert api.get_person(id=invitation.created_by).email == "test@tester.com"


def test_add_invitation_no_created_by(api, dummy_badge_id, dummy_person_id):
    with pytest.raises(ValueError):
        api.add_invitation(dummy_badge_id)


def test_last_login(api, callback_calls, dummy_person_id):
    person = api.get_person(dummy_person_id)
    assert not person.last_login
    api.note_login(nickname=person.nickname)
    assert person.last_login

    assert len(callback_calls) == 1
    message = callback_calls[0][0][0]
    assert message.body == {"user": {"username": "test", "badges_user_id": 1}}
    assert message.agent_name == "test"
    assert message.summary == "test logged into badges for the first time"


def test_add_assertion(api, callback_calls, dummy_badge_id, dummy_person_id):
    api.add_assertion(dummy_badge_id, "test@tester.com", None, "link")
    assert api.assertion_exists(dummy_badge_id, "test@tester.com")

    badge = api.get_badge(dummy_badge_id)
    assert badge.assertions[0].issued_for == "link"
    assert api.get_assertions_by_badge(dummy_badge_id)[0].__str__() == "TestBadge<->test@tester.com"

    # Ensure that we would have published two fedmsg messages for that.
    assert len(callback_calls) == 2

    award_message = callback_calls[0][0][0]
    rank_advance_message = callback_calls[1][0][0]

    assert award_message.body == {
        "badge": {
            "name": "TestBadge",
            "description": "A test badge for doing unit tests",
            "image_url": "TestImage",
            "badge_id": "testbadge",
        },
        "user": {"username": "test", "badges_user_id": 1},
    }
    assert award_message.agent_name == "test"
    assert award_message.summary == "test was awarded the badge `TestBadge`"

    assert rank_advance_message.body == {
        "person": {
            "email": "test@tester.com",
            "id": 1,
            "nickname": "test",
            "website": None,
            "bio": None,
            "rank": 1,
        },
        "old_rank": None,
    }
    assert rank_advance_message.agent_name == "test"
    assert rank_advance_message.summary == "test's Badges rank changed from None to 1"


@pytest.mark.parametrize("test_email", ["test@tester.com", "Test@Tester.Com"])
def test_get_assertions_by_email(api, callback_calls, dummy_badge_id, dummy_person_id, test_email):
    api.add_assertion(dummy_badge_id, "test@tester.com", None, "link")
    # This should be case-insensitive
    assertions = api.get_assertions_by_email(person_email=test_email)
    assert len(assertions) == 1
    assert str(assertions[0]) == "TestBadge<->test@tester.com"


def test_get_badges_from_tags(api, dummy_issuer_id):
    # Badge tagged with "test"
    api.add_badge(
        "TestBadgeA",
        "TestImage",
        "A test badge for doing unit tests",
        "TestCriteria",
        dummy_issuer_id,
        tags="test",
    )

    # Badge tagged with "tester"
    api.add_badge(
        "TestBadgeB",
        "TestImage",
        "A second test badge for doing unit tests",
        "TestCriteria",
        dummy_issuer_id,
        tags="tester",
    )

    # Badge tagged with both "test" and "tester"
    api.add_badge(
        "TestBadgeC",
        "TestImage",
        "A third test badge for doing unit tests",
        "TestCriteria",
        dummy_issuer_id,
        tags="test, tester",
    )

    tags = ["test", "tester"]
    badges_any = api.get_badges_from_tags(tags, match_all=False)
    assert len(badges_any) == 3
    badges_all = api.get_badges_from_tags(tags, match_all=True)
    assert len(badges_all) == 1



"""Add tags field to badges.

Revision ID: 5791a2b9fb6a
Revises: 39583332f4ea
Create Date: 2013-06-10 13:05:40.130328

"""

import sqlalchemy as sa
from alembic import op


# revision identifiers, used by Alembic.
revision = "5791a2b9fb6a"
down_revision = "39583332f4ea"


def upgrade():
    op.add_column("badges", sa.Column("tags", sa.Text))


def downgrade():
    # Downgrade will fail if using SQLite. The transaction will
    # be rolled back as per env.py.
    op.drop_column("badges", "tags")



"""Add profile fields to persons.

Revision ID: 39583332f4ea
Revises: fa1d309e8c3
Create Date: 2013-06-10 12:33:45.319780

"""

import sqlalchemy as sa
from alembic import op


# revision identifiers, used by Alembic.
revision = "39583332f4ea"
down_revision = "fa1d309e8c3"


def upgrade():
    op.add_column("persons", sa.Column("nickname", sa.Text))
    op.add_column("persons", sa.Column("website", sa.Text))
    op.add_column("persons", sa.Column("bio", sa.Text))


def downgrade():
    # Downgrade will fail if using SQLite. The transaction will
    # be rolled back as per env.py.
    op.drop_column("persons", "nickname")
    op.drop_column("persons", "website")
    op.drop_column("persons", "bio")



"""Add an stl column for badges.

Revision ID: 508367dcbbb5
Revises: 2879ed5a6297
Create Date: 2014-07-11 09:36:33.211281

"""

import sqlalchemy as sa
from alembic import op


# revision identifiers, used by Alembic.
revision = "508367dcbbb5"
down_revision = "2879ed5a6297"


def upgrade():
    op.add_column("badges", sa.Column("stl", sa.Unicode(128)))


def downgrade():
    # Downgrade will fail if using SQLite. The transaction will
    # be rolled back as per env.py.
    op.drop_column("badges", "stl")



"""CurrentValue

Revision ID: 51261da641fb
Revises: 3d3fb9e59e7b
Create Date: 2024-07-23 15:27:48.303528
"""

import sqlalchemy as sa
from alembic import op


# revision identifiers, used by Alembic.
revision = "51261da641fb"
down_revision = "3d3fb9e59e7b"
branch_labels = None
depends_on = None


def upgrade():
    op.create_table(
        "current_values",
        sa.Column("badge_id", sa.Unicode(length=128), nullable=False),
        sa.Column("person_id", sa.Integer(), nullable=False),
        sa.Column("value", sa.Integer(), nullable=False),
        sa.Column("last_update", sa.DateTime(), nullable=False),
        sa.ForeignKeyConstraint(
            ["badge_id"], ["badges.id"], name=op.f("fk_current_values_badge_id_badges")
        ),
        sa.ForeignKeyConstraint(
            ["person_id"], ["persons.id"], name=op.f("fk_current_values_person_id_persons")
        ),
        sa.PrimaryKeyConstraint("badge_id", "person_id", name=op.f("pk_current_values")),
    )


def downgrade():
    op.drop_table("current_values")



"""Add Person.avatar

Revision ID: 3d3fb9e59e7b
Revises: e7040e76728
Create Date: 2024-05-17 13:47:31.271833
"""

import sqlalchemy as sa
from alembic import op


# revision identifiers, used by Alembic.
revision = "3d3fb9e59e7b"
down_revision = "e7040e76728"
branch_labels = None
depends_on = None


def upgrade():
    op.add_column("persons", sa.Column("_avatar", sa.Unicode(length=128), nullable=True))


def downgrade():
    op.drop_column("persons", "_avatar")



"""Add created_on field to badges table.

Revision ID: 420c02357a1b
Revises: None
Create Date: 2013-06-09 22:23:32.519635

"""

import sqlalchemy as sa
from alembic import op


# revision identifiers, used by Alembic.
revision = "420c02357a1b"
down_revision = None


def upgrade():
    op.add_column("badges", sa.Column("created_on", sa.DateTime, nullable=False))


def downgrade():
    # This will fail if using SQLite. The transaction will
    # be rolled back.
    op.drop_column("badges", "created_on")



import re

import click

from ..model import Badge, Milestone, Series
from .utils import get_db_manager_from_config


_ROMAN_TO_ARABIC = dict(
    [("I", 1), ("V", 5), ("X", 10), ("L", 50), ("C", 100), ("D", 500), ("M", 1000)]
)
_REPLACEMENTS = [
    ("CM", "DCCCC"),
    ("CD", "CCCC"),
    ("XC", "LXXXX"),
    ("XL", "XXXX"),
    ("IX", "VIIII"),
    ("IV", "IIII"),
]
# A name of a badge in series must end with parenthesised series name and
# ordinal number, either in arabic or roman numerals. All badges in a given
# series must share the same series name.
_SERIES_NAME_RE = re.compile(r".+ \((?P<name>.+) (?P<ord>[0-9IXVL]+)\)")


def _convert(mapping, x):
    for prefix, replacement in mapping:
        x = x.replace(prefix, replacement)
    return x


def _to_number(x):
    """Convert a string with Roman numerals into an integer."""
    total = 0
    for c in _convert(_REPLACEMENTS, x):
        total += _ROMAN_TO_ARABIC[c]
    return total


def get_series_name(name):
    """Given a badge name, return a tuple of series name and ordinal number of
    this badge in the series.

    If the badge is not in any series, both tuple elements are None.
    """
    m = _SERIES_NAME_RE.match(name)
    if not m:
        return None, None
    base = m.group("name")
    idx = m.group("ord")
    try:
        try:
            return base, int(idx)
        except ValueError:
            return base, _to_number(idx)
    except (ValueError, KeyError):
        return None, None


@click.command()
@click.argument("config", type=click.Path(exists=True))
def main(config):
    db_mgr = get_db_manager_from_config(config)
    with db_mgr.Session() as session:
        for badge in session.query(Badge).all():
            if badge.milestone:
                # Skip badges that already are in some series.
                continue
            series_name, ordering = get_series_name(badge.name)
            if series_name and ordering:
                series = session.query(Series).filter(Series.name == series_name).first()

                if not series:
                    print(
                        f"Series <{series_name}> does not exist, skipping "
                        f"processing badge {badge.name}"
                    )
                    continue
                milestone = Milestone()
                milestone.badge_id = badge.id
                milestone.position = ordering
                milestone.series_id = series_name.lower().replace(" ", "-")
                session.add(milestone)
        session.commit()
